"""
@file
@brief About completion
"""
from typing import Tuple, List, Iterator
from collections import deque


class CompletionTrieNode:
    """
    Node definition in a trie used to do completion, see :ref:`l-completion0`.
    This implementation is not very efficient about memmory consumption,
    it does not hold above 200.000 words.
    It should be done another way (:epkg:`cython`, :epkg:`C++`).
    """

    __slots__ = ("value", "children", "weight",
                 "leave", "stat", "parent", "disp")

    def __init__(self, value, leave, weight=1.0, disp=None):
        """
        @param      value       value (a character)
        @param      leave       boolean (is it a completion)
        @param      weight      ordering (the lower, the first)
        @param      disp        original string, use this to identify the node
        """
        if not isinstance(value, str):
            raise TypeError(
                "value must be str not '{0}' - type={1}".format(value, type(value)))
        self.value = value
        self.children = None
        self.weight = weight
        self.leave = leave
        self.stat = None
        self.parent = None
        self.disp = disp

    @property
    def root(self):
        """
        Returns the initial node with no parent.
        """
        node = self
        while node.parent is not None:
            node = node.parent
        return node

    def __str__(self):
        """
        usual
        """
        return "[{2}:{0}:w={1}]".format(self.value, self.weight, "#" if self.leave else "-")

    def _add(self, key, child):
        """
        Adds a child.

        @param      key         one letter of the word
        @param      child       child
        @return                 self
        """
        if self.children is None:
            self.children = {key: child}
            child.parent = self
        elif key in self.children:
            raise KeyError("'{0}' already added".format(key))
        else:
            self.children[key] = child
            child.parent = self
        return self

    def items_list(self) -> List['CompletionTrieNode']:
        """
        All children nodes inluding itself in a list.

        @return          list[
        """
        res = [self]
        if self.children is not None:
            for _, v in sorted(self.children.items()):
                r = v.items_list()
                res.extend(r)
        return res

    def __iter__(self):
        """
        Iterates on all nodes (sorted).
        """
        stack = [self]
        while len(stack) > 0:
            node = stack.pop()
            yield node
            if node.children:
                stack.extend(v for k, v in sorted(
                    node.children.items(), reverse=True))

    def unsorted_iter(self):
        """
        Iterates on all nodes.
        """
        stack = [self]
        while len(stack) > 0:
            node = stack.pop()
            yield node
            if node.children:
                stack.extend(node.children.values())

    def items(self) -> Iterator[Tuple[float, str, 'CompletionTrieNode']]:
        """
        Iterates on children, iterates on weight, key, child.
        """
        if self.children is not None:
            for k, v in self.children.items():
                yield v.weight, k, v

    def iter_leaves(self, max_weight=None) -> Iterator[Tuple[float, str]]:
        """
        Iterators on leaves sorted per weight, yield weight, value.

        @param      max_weight  keep all value under this threshold or None for all
        """
        def iter_local(node):
            if node.leave and (max_weight is None or node.weight <= max_weight):
                yield node.weight, None, node.value
            for w, k, v in sorted(node.items()):
                for w_, k_, v_ in iter_local(v):
                    yield w_, k_, v_

        for w, _, v in sorted(iter_local(self)):
            yield w, v

    def leaves(self) -> Iterator['CompletionTrieNode']:
        """
        Iterators on leaves.
        """
        stack = [self]
        while len(stack) > 0:
            pop = stack.pop()
            if pop.leave:
                yield pop
            if pop.children:
                stack.extend(pop.children.values())

    def all_completions(self) -> List[Tuple['CompletionTrieNone', List[str]]]:
        """
        Retrieves all completions for a node,
        the method does not need @see me precompute_stat to be run first.
        """
        word = self.value
        nodes = [self.root]
        node = nodes[0]
        for c in word:
            if node.children is not None and c in node.children:
                node = node.children[c]
                nodes.append(node)
        nodes.reverse()
        all_res = []
        for node in nodes:
            res = list(n[1] for n in node.iter_leaves())
            all_res.append((node, res))
        all_res.reverse()
        return all_res

    def all_mks_completions(self) -> List[Tuple['CompletionTrieNone', List['CompletionTrieNone']]]:
        """
        Retrieves all completions for a node,
        the method assumes @see me precompute_stat was run.
        """
        res = []
        node = self
        while True:
            res.append((node, node.stat.completions))
            if node.parent is None:
                break
            node = node.parent
        res.reverse()
        return res

    def str_all_completions(self, maxn=10, use_precompute=True) -> str:
        """
        Builds a string with all completions for all
        prefixes along the paths.

        @param      maxn            maximum number of completions to show
        @param      use_precompute  use intermediate results built by @see me precompute_stat
        @return                     str
        """
        res = self.all_mks_completions() if use_precompute else self.all_completions()
        rows = []
        for node, sug in res:
            rows.append("l={3} p='{0}' {1} {2}".format(node.value, "-" * 10, node.stat.str_mks(),
                                                       '+' if node.leave else '-'))
            for i, s in enumerate(sug):
                if isinstance(s, str):
                    rows.append("  {0}-'{1}'".format(i + 1, s))
                else:
                    rows.append(
                        "  {0}-w{1}-'{2}'".format(i + 1, s[0], s[1].value))
                if maxn is not None and i > maxn:
                    break
        return "\n".join(rows)

    @staticmethod
    def build(words) -> 'CompletionTrieNode':
        """
        Builds a trie.

        @param  words       list of ``(word)`` or ``(weight, word)`` or ``(weight, word, display string)``
        @return             root of the trie (CompletionTrieNode)
        """
        root = CompletionTrieNode('', False)
        nb = 0
        minw = None
        for wword in words:
            if isinstance(wword, tuple):
                if len(wword) == 2:
                    w, word = wword
                    disp = None
                elif len(wword) == 3:
                    w, word, disp = wword
                else:
                    raise ValueError(
                        "Unexpected number of values, it should be (weight, word) or (weight, word, dispplay string): {0}".format(wword))
            else:
                w = 1.0
                word = wword
                disp = None
            if w is None:
                w = nb
            if minw is None or minw > w:
                minw = w
            node = root
            new_node = None
            for c in word:
                if node.children is not None and c in node.children:
                    if not node.leave:
                        node.weight = min(node.weight, w)
                    node = node.children[c]
                else:
                    new_node = CompletionTrieNode(
                        node.value + c, False, weight=w)
                    node._add(c, new_node)
                    node = new_node
            if new_node is None:
                if node.leave:
                    raise ValueError(
                        "Value '{0}' appears twice in the input list (not allowed).".format(word))
                new_node = node
            new_node.leave = True
            new_node.weight = w
            if disp is not None:
                new_node.disp = disp
            nb += 1
        root.weight = minw
        return root

    def find(self, prefix: str) -> 'CompletionTrieNode':
        """
        Returns the node which holds all completions starting with a given prefix.

        @param      prefix      prefix
        @return                 node or None for no result
        """
        if len(prefix) == 0:
            if not self.value:
                return self
            else:
                raise ValueError(
                    "find '{0}' but node is not empty '{1}'".format(prefix, self.value))
        node = self
        for c in prefix:
            if node.children is not None and c in node.children:
                node = node.children[c]
            else:
                return None
        return node

    def min_keystroke(self, word: str) -> Tuple[int, int]:
        """
        Returns the minimum keystrokes for a word without optimisation,
        this function should be used if you only have a couple of values to
        computes. You shoud use @see me min_keystroke0 to compute all of them.

        @param      word        word
        @return                 number, length of best prefix

        See :ref:`l-completion-optim`.

        .. math::
            :nowrap:

            \\begin{eqnarray*}
            K(q, k, S) &=& \\min\\acc{ i | s_i \\succ q[1..k], s_i \\in S } \\\\
            M(q, S) &=& \\min_{0 \\infegal k \\infegal l(q)}  k + K(q, k, S)
            \\end{eqnarray*}
        """
        nodes = [self]
        node = self
        for c in word:
            if node.children is not None and c in node.children:
                node = node.children[c]
                nodes.append(node)
            else:
                # not found
                return len(word), -1
        nodes.reverse()
        metric = len(word)
        best = len(word)
        for node in nodes[1:]:
            res = list(n[1] for n in node.iter_leaves())
            ind = res.index(word)
            m = len(node.value) + ind + 1
            if m < metric:
                metric = m
                best = len(node.value)
            if ind >= len(word):
                # no need to go further, the position will increase
                break
        return metric, best

    def min_keystroke0(self, word: str) -> Tuple[int, int]:
        """
        Returns the minimum keystrokes for a word.

        @param      word        word
        @return                 number, length of best prefix, iteration it stops moving

        This function must be called after @see me precompute_stat
        and @see me update_stat_dynamic.

        See :ref:`l-completion-optim`.

        .. math::
            :nowrap:

            \\begin{eqnarray*}
            K(q, k, S) &=& \\min\\acc{ i | s_i \\succ q[1..k], s_i \\in S } \\\\
            M(q, S) &=& \\min_{0 \\infegal k \\infegal l(q)}  k + K(q, k, S)
            \\end{eqnarray*}
        """
        node = self.find(word)
        if node is None:
            raise NotImplementedError(
                "this metric is not yet computed for a query outside the trie: '{0}'".format(word))
        if not hasattr(node, "stat"):
            raise AttributeError("run precompute_stat and update_stat_dynamic")
        if not hasattr(node.stat, "mks1"):
            raise AttributeError("run precompute_stat and update_stat_dynamic\nnode={0}\n{1}".format(
                self, "\n".join(sorted(self.stat.__dict__.keys()))))
        return node.stat.mks0, node.stat.mks0_, 0

    def min_dynamic_keystroke(self, word: str) -> Tuple[int, int]:
        """
        Returns the dynamic minimum keystrokes for a word.

        @param      word        word
        @return                 number, length of best prefix, iteration it stops moving

        This function must be called after @see me precompute_stat
        and @see me update_stat_dynamic.
        See :ref:`Dynamic Minimum Keystroke <def-mks2>`.

        .. math::
            :nowrap:

            \\begin{eqnarray*}
            K(q, k, S) &=& \\min\\acc{ i | s_i \\succ q[1..k], s_i \\in S } \\\\
            M'(q, S) &=& \\min_{0 \\infegal k \\infegal l(q)} \\acc{ M'(q[1..k], S) + K(q, k, S) | q[1..k] \\in S }
            \\end{eqnarray*}
        """
        node = self.find(word)
        if node is None:
            raise NotImplementedError(
                "this metric is not yet computed for a query outside the trie: '{0}'".format(word))
        if not hasattr(node, "stat"):
            raise AttributeError("run precompute_stat and update_stat_dynamic")
        if not hasattr(node.stat, "mks1"):
            raise AttributeError("run precompute_stat and update_stat_dynamic\nnode={0}\n{1}".format(
                self, "\n".join(sorted(self.stat.__dict__.keys()))))
        return node.stat.mks1, node.stat.mks1_, node.stat.mks1i_

    def min_dynamic_keystroke2(self, word: str) -> Tuple[int, int]:
        """
        Returns the modified dynamic minimum keystrokes for a word.

        @param      word        word
        @return                 number, length of best prefix, iteration it stops moving

        This function must be called after @see me precompute_stat
        and @see me update_stat_dynamic.
        See :ref:`Modified Dynamic Minimum Keystroke <def-mks3>`.

        .. math::
            :nowrap:

            \\begin{eqnarray*}
            K(q, k, S) &=& \\min\\acc{ i | s_i \\succ q[1..k], s_i \\in S } \\\\
            M"(q, S) &=& \\min \\left\\{ \\begin{array}{l}
                            \\min_{1 \\infegal k \\infegal l(q)} \\acc{ M"(q[1..k-1], S) + 1 + K(q, k, S) | q[1..k] \\in S } \\\\
                            \\min_{0 \\infegal k \\infegal l(q)} \\acc{ M"(q[1..k], S) + \\delta + K(q, k, S) | q[1..k] \\in S }
                            \\end{array} \\right .
            \\end{eqnarray*}
        """
        node = self.find(word)
        if node is None:
            raise NotImplementedError(
                "this metric is not yet computed for a query outside the trie: '{0}'".format(word))
        if not hasattr(node, "stat"):
            raise AttributeError("run precompute_stat and update_stat_dynamic")
        if not hasattr(node.stat, "mks2"):
            raise AttributeError("run precompute_stat and update_stat_dynamic\nnode={0}\n{1}".format(
                self, "\n".join(sorted(self.stat.__dict__.keys()))))
        return node.stat.mks2, node.stat.mks2_, node.stat.mks2i_

    def precompute_stat(self):
        """
        Computes and stores list of completions for each node,
        computes *mks*.

        @param      clean   clean stat
        """
        stack = deque()
        stack.extend(self.leaves())
        while len(stack) > 0:
            pop = stack.popleft()
            if pop.stat is not None:
                continue
            if not pop.children:
                pop.stat = CompletionTrieNode._Stat()
                pop.stat.completions = []
                pop.stat.mks0 = len(pop.value)
                pop.stat.mks0_ = len(pop.value)
                if pop.parent is not None:
                    stack.append(pop.parent)
            elif all(v.stat is not None for v in pop.children.values()):
                pop.stat = CompletionTrieNode._Stat()
                if pop.leave:
                    pop.stat.mks0 = len(pop.value)
                    pop.stat.mks0_ = len(pop.value)
                stack.extend(pop.children.values())
                pop.stat.merge_completions(pop.value, pop.children.values())
                pop.stat.next_nodes = pop.children
                pop.stat.update_minimum_keystroke(len(pop.value))
                if pop.parent is not None:
                    stack.append(pop.parent)
            else:
                # we'll do it again later
                stack.append(pop)

    def update_stat_dynamic(self, delta=0.8):
        """
        Must be called after @see me precompute_stat
        and computes dynamic mks (see :ref:`Dynamic Minimum Keystroke <def-mks2>`).

        @param      delta       parameter :math:`\\delta` in defintion
                                :ref:`Modified Dynamic KeyStroke <def-mks3>`
        @return                 number of iterations to converge
        """
        for node in self.unsorted_iter():
            node.stat.init_dynamic_minimum_keystroke(len(node.value))
            node.stat.iter_ = 0
        updates = 1
        itera = 0
        while updates > 0:
            updates = 0
            stack = []
            stack.append(self)
            while len(stack) > 0:
                pop = stack.pop()
                if pop.stat.iter_ > itera:
                    continue
                updates += pop.stat.update_dynamic_minimum_keystroke(
                    len(pop.value), delta)
                if pop.children:
                    stack.extend(pop.children.values())
                pop.stat.iter_ += 1
            itera += 1
        return itera

    ##
    # end of methods, beginning of subclasses
    ##

    class _Stat:
        """
        Stores statistics and intermediate data about the compuation the metrics.

        It contains the following members:

        * mks0*: value of minimum keystroke
        * mks0_*: length of the prefix to obtain *mks0*
        * *mks_iter*: current iteration during the computation of mks

        * *mks1*: value of dynamic minimum keystroke
        * *mks1_*: length of the prefix to obtain *mks*
        * *mks1i_*: iteration when it was obtained

        * *mks2*: value of modified dynamic minimum keystroke
        * *mks2_*: length of the prefix to obtain *mks2*
        * *mks2i*: iteration when it converged
        """

        def merge_completions(self, prefix: int, nodes: '[CompletionTrieNode]'):
            """
            Merges list of completions and cut the list, we assume
            given lists are sorted.
            """
            class Fake:
                pass
            res = []
            indexes = [0 for _ in nodes]
            indexes.append(0)
            last = Fake()
            last.value = None
            last.stat = CompletionTrieNode._Stat()
            last.stat.completions = list(
                sorted((_.weight, _) for _ in nodes if _.leave))
            nodes = list(nodes)
            nodes.append(last)

            maxl = 0
            while True:
                en = [(_.stat.completions[indexes[i]][0], i, _.stat.completions[indexes[i]][1])
                      for i, _ in enumerate(nodes) if indexes[i] < len(_.stat.completions)]
                if not en:
                    break
                e = min(en)
                i = e[1]
                res.append((e[0], e[2]))
                indexes[i] += 1
                maxl = max(maxl, len(res[-1][1].value))

            # maxl - len(prefix) represents the longest list which reduces the number of keystrokes
            # however, as the method aggregates completions at a lower lovel,
            # we must keep longer completions for lower levels
            ind = maxl
            if len(res) > ind:
                self.completions = res[:ind]
            else:
                self.completions = res

        def update_minimum_keystroke(self, lw):
            """
            Updates minimum keystroke for the completions.

            @param      lw      prefix length
            """
            for i, wsug in enumerate(self.completions):
                sug = wsug[1]
                nl = lw + i + 1
                if not hasattr(sug.stat, "mks0") or sug.stat.mks0 > nl:
                    sug.stat.mks0 = nl
                    sug.stat.mks0_ = lw

        def update_dynamic_minimum_keystroke(self, lw, delta):
            """
            Updates dynamic minimum keystroke for the completions.

            @param      lw      prefix length
            @param      delta   parameter :math:`\\delta` in defintion
                                :ref:`Modified Dynamic KeyStroke <def-mks3>`
            @return             number of updates
            """
            self.mks_iter += 1
            update = 0
            for i, wsug in enumerate(self.completions):
                sug = wsug[1]
                if sug.leave:
                    # this is a leave so we consider the completion being part
                    # of the list of completions
                    nl = self.mks1 + i + 1
                    if sug.stat.mks1 > nl:
                        sug.stat.mks1 = nl
                        sug.stat.mks1_ = lw
                        sug.stat.mks1i_ = self.mks_iter
                        update += 1
                    nl = self.mks2 + i + 1 + delta
                    if sug.stat.mks2 > nl:
                        sug.stat.mks2 = nl
                        sug.stat.mks2_ = lw
                        sug.stat.mks2i_ = self.mks_iter
                        update += 1
                else:
                    raise Exception("this case should not happen")

            # optimisation of second case of modified metric
            # in a separate function for profiling
            def second_step(update):
                if hasattr(self, "next_nodes"):
                    for _, child in self.next_nodes.items():
                        for i, wsug in enumerate(child.stat.completions):
                            sug = wsug[1]
                            if not sug.leave:
                                continue
                            nl = self.mks2 + i + 2
                            if sug.stat.mks2 > nl:
                                sug.stat.mks2 = nl
                                sug.stat.mks2_ = lw
                                sug.stat.mks2i_ = self.mks_iter
                                update += 1
                return update

            update = second_step(update)

            # finally we need to update mks, mks2 for every prefix
            # this is not necessary a leave so it does not appear in the list of completions
            # but we need to update mks for these strings, we assume it just
            # requires an extra character, somehow, we propagate the values
            if hasattr(self, "next_nodes"):
                for _, n in self.next_nodes.items():
                    if not hasattr(n.stat, "mks1") or n.stat.mks1 > self.mks1 + 1:
                        n.stat.mks1 = self.mks1 + 1
                        n.stat.mks1_ = self.mks1_
                        n.stat.mks1i_ = self.mks_iter
                        update += 1
                    if not hasattr(n.stat, "mks2") or n.stat.mks2 > self.mks2 + 1:
                        n.stat.mks2 = self.mks2 + 1
                        n.stat.mks2_ = self.mks2_
                        n.stat.mks2i_ = self.mks_iter
                        update += 1

            return update

        def init_dynamic_minimum_keystroke(self, lw):
            """
            Initializes *mks* and *mks2* from from *mks0*.

            @param      lw      length of the prefix
            """
            if hasattr(self, "mks0"):
                self.mks1 = self.mks0
                self.mks1_ = self.mks0_
                self.mks_iter = 0
                self.mks1i_ = 0
                self.mks2 = self.mks0
                self.mks2_ = self.mks0_
                self.mks2i_ = 0
            else:
                self.mks0 = lw
                self.mks0_ = 0
                self.mks1 = lw
                self.mks1_ = lw
                self.mks_iter = 0
                self.mks1i_ = 0
                self.mks2 = lw
                self.mks2_ = lw
                self.mks2i_ = 0

        def str_mks0(self) -> str:
            """
            Returns a string with metric information.
            """
            if hasattr(self, "mks0"):
                return "MKS={0} *={1}".format(self.mks0, self.mks0_)
            else:
                return "-"

        def str_mks(self) -> str:
            """
            Returns a string with metric information.
            """
            s0 = self.str_mks0()
            if hasattr(self, "mks1"):
                return s0 + " |'={0} *={1},{2} |\"={3} *={4},{5} |nn={6}".format(
                    self.mks1, self.mks1_, self.mks1i_, self.mks2, self.mks2i_, self.mks2i_, '+' if hasattr(self, "next_nodes") else '-')
            else:
                return s0



# -*- coding: utf-8 -*-
"""
@file
@brief About Machine Learning Benchmark
"""
import os
import numpy
from sklearn.model_selection import train_test_split
from sklearn.base import ClusterMixin
from sklearn.metrics import silhouette_score
from pyquickhelper.loghelper import noLOG
from pyquickhelper.benchhelper import GridBenchMark


class MlGridBenchMark(GridBenchMark):
    """
    The class tests a list of model over a list of datasets.
    """

    def __init__(self, name, datasets, clog=None, fLOG=noLOG, path_to_images=".",
                 cache_file=None, progressbar=None, graphx=None, graphy=None,
                 **params):
        """
        @param      name            name of the test
        @param      datasets        list of dictionary of dataframes
        @param      clog            see @see cl CustomLog or string
        @param      fLOG            logging function
        @param      params          extra parameters
        @param      path_to_images  path to images and intermediate results
        @param      cache_file      cache file
        @param      progressbar     relies on *tqdm*, example *tnrange*
        @param      graphx          list of variables to use as X axis
        @param      graphy          list of variables to use as Y axis

        If *cache_file* is specified, the class will store the results of the
        method :meth:`bench <pyquickhelper.benchhelper.benchmark.GridBenchMark.bench>`.
        On a second run, the function load the cache
        and run modified or new run (in *param_list*).

        *datasets* should be a dictionary with dataframes a values
        with the following keys:

        * ``'X'``: features
        * ``'Y'``: labels (optional)
        """
        GridBenchMark.__init__(self, name=name, datasets=datasets, clog=clog, fLOG=fLOG,
                               path_to_images=path_to_images, cache_file=cache_file,
                               progressbar=progressbar, **params)
        self._xaxis = graphx
        self._yaxis = graphy

    def preprocess_dataset(self, dsi, **params):
        """
        Splits the dataset into train and test.

        @param      dsi         dataset index
        @param      params      additional parameters
        @return                 dataset (like info), dictionary for metrics
        """
        ds, appe, params = GridBenchMark.preprocess_dataset(
            self, dsi, **params)

        no_split = ds["no_split"] if "no_split" in ds else False

        if no_split:
            self.fLOG("[MlGridBenchMark.preprocess_dataset] no split")
            return (ds, ds), appe, params

        self.fLOG("[MlGridBenchMark.preprocess_dataset] split train test")
        spl = ["X", "Y", "weight", "group"]
        names = [_ for _ in spl if _ in ds]
        if len(names) == 0:
            raise ValueError(  # pragma: no cover
                "No dataframe or matrix was found.")
        mats = [ds[_] for _ in names]

        pars = {"train_size", "test_size"}
        options = {k: v for k, v in params.items() if k in pars}
        for k in pars:
            if k in params:
                del params[k]

        res = train_test_split(*mats, **options)

        train = {}
        for i, n in enumerate(names):
            train[n] = res[i * 2]
        test = {}
        for i, n in enumerate(names):
            test[n] = res[i * 2 + 1]

        self.fLOG("[MlGridBenchMark.preprocess_dataset] done")
        return (train, test), appe, params

    def bench_experiment(self, ds, **params):  # pylint: disable=W0237
        """
        Calls meth *fit*.
        """
        if not isinstance(ds, tuple) and len(ds) != 2:
            raise TypeError(  # pragma: no cover
                "ds must a tuple with two dictionaries train, test")
        if "model" not in params:
            raise KeyError(  # pragma: no cover
                "params must contains key 'model'")
        model = params["model"]
        # we assume model is a function which creates a model
        model = model()
        del params["model"]
        return self.fit(ds[0], model, **params)

    def predict_score_experiment(self, ds, model, **params):  # pylint: disable=W0237
        """
        Calls method *score*.
        """
        if not isinstance(ds, tuple) and len(ds) != 2:
            raise TypeError(  # pragma: no cover
                "ds must a tuple with two dictionaries train, test")
        if "model" in params:
            raise KeyError(  # pragma: no cover
                "params must not contains key 'model'")
        return self.score(ds[1], model, **params)

    def fit(self, ds, model, **params):
        """
        Trains a model.

        @param      ds          dictionary with the data to use for training
        @param      model       model to train
        """
        if "X" not in ds:
            raise KeyError(  # pragma: no cover
                "ds must contain key 'X'")
        if "model" in params:
            raise KeyError(  # pragma: no cover
                "params must not contain key 'model', this is the model to train")
        X = ds["X"]
        Y = ds.get("Y", None)
        weight = ds.get("weight", None)
        self.fLOG("[MlGridBenchMark.fit] fit", params)

        train_params = params.get("train_params", {})

        if weight is not None:
            model.fit(X=X, y=Y, weight=weight, **train_params)
        else:
            model.fit(X=X, y=Y, **train_params)
        self.fLOG("[MlGridBenchMark.fit] Done.")
        return model

    def score(self, ds, model, **params):
        """
        Scores a model.
        """
        X = ds["X"]
        Y = ds.get("Y", None)

        if "weight" in ds:
            raise NotImplementedError(  # pragma: no cover
                "weight are not used yet")

        metrics = {}
        appe = {}

        if hasattr(model, "score"):
            score = model.score(X, Y)
            metrics["own_score"] = score

        if isinstance(model, ClusterMixin):
            # add silhouette
            if hasattr(model, "predict"):
                ypred = model.predict(X)
            elif hasattr(model, "transform"):
                ypred = model.transform(X)
            elif hasattr(model, "labels_"):
                ypred = model.labels_
            if len(ypred.shape) > 1 and ypred.shape[1] > 1:
                ypred = numpy.argmax(ypred, axis=1)
            score = silhouette_score(X, ypred)
            metrics["silhouette"] = score

        return metrics, appe

    def end(self):
        """
        nothing to do
        """
        pass

    def graphs(self, path_to_images):
        """
        Plots multiples graphs.

        @param      path_to_images  where to store images
        @return     list of tuple (image_name, function to create the graph)
        """
        import matplotlib.pyplot as plt  # pylint: disable=C0415
        import matplotlib.cm as mcm  # pylint: disable=C0415
        df = self.to_df()

        def local_graph(vx, vy, ax=None, text=True, figsize=(5, 5)):
            btrys = set(df["_btry"])
            ymin = df[vy].min()
            ymax = df[vy].max()
            decy = (ymax - ymin) / 50
            colors = mcm.rainbow(numpy.linspace(0, 1, len(btrys)))
            if len(btrys) == 0:
                raise ValueError("The benchmark is empty.")  # pragma: no cover
            if ax is None:
                _, ax = plt.subplots(1, 1, figsize=figsize)  # pragma: no cover
                ax.grid(True)  # pragma: no cover
            for i, btry in enumerate(sorted(btrys)):
                subset = df[df["_btry"] == btry]
                if subset.shape[0] > 0:
                    tx = subset[vx].mean()
                    ty = subset[vy].mean()
                    if not numpy.isnan(tx) and not numpy.isnan(ty):
                        subset.plot(x=vx, y=vy, kind="scatter",
                                    label=btry, ax=ax, color=colors[i])
                        if text:
                            ax.text(tx, ty + decy, btry, size='small',
                                    color=colors[i], ha='center', va='bottom')
            ax.set_xlabel(vx)
            ax.set_ylabel(vy)
            return ax

        res = []
        if self._xaxis is not None and self._yaxis is not None:
            for vx in self._xaxis:
                for vy in self._yaxis:
                    self.fLOG("Plotting {0} x {1}".format(vx, vy))
                    func_graph = lambda ax=None, text=True, vx=vx, vy=vy, **kwargs: \
                        local_graph(vx, vy, ax=ax, text=text, **kwargs)

                    if path_to_images is not None:
                        img = os.path.join(
                            path_to_images, "img-{0}-{1}x{2}.png".format(self.Name, vx, vy))
                        gr = self.LocalGraph(
                            func_graph, img, root=path_to_images)
                        self.fLOG("Saving '{0}'".format(img))
                        fig, ax = plt.subplots(1, 1, figsize=(8, 8))
                        gr.plot(ax=ax, text=True)
                        fig.savefig(img)
                        self.fLOG("Done")
                        res.append(gr)
                        plt.close('all')
                    else:
                        gr = self.LocalGraph(func_graph)
                        res.append(gr)
        return res

    def plot_graphs(self, grid=None, text=True, **kwargs):
        """
        Plots all graphs in the same graphs.

        @param      grid        grid of axes
        @param      text        add legend title on the graph
        @return                 grid
        """
        nb = len(self.Graphs)
        if nb == 0:
            raise ValueError("No graph to plot.")  # pragma: no cover

        nb = len(self.Graphs)
        if nb % 2 == 0:
            size = nb // 2, 2
        else:
            size = nb // 2 + 1, 2

        if grid is None:
            import matplotlib.pyplot as plt  # pylint: disable=C0415
            fg = kwargs.get('figsize', (5 * size[0], 10))
            _, grid = plt.subplots(size[0], size[1], figsize=fg)
            if 'figsize' in kwargs:
                del kwargs['figsize']  # pragma: no cover
        else:
            shape = grid.shape
            if shape[0] * shape[1] < nb:
                raise ValueError(  # pragma: no cover
                    "The graph is not big enough {0} < {1}".format(shape, nb))

        x = 0
        y = 0
        for i, gr in enumerate(self.Graphs):
            self.fLOG("Plot graph {0}/{1}".format(i + 1, nb))
            gr.plot(ax=grid[y, x], text=text, **kwargs)
            x += 1
            if x >= grid.shape[1]:
                x = 0
                y += 1
        return grid



# -*- coding: utf-8 -*-
"""
@file
@brief Détecte les segments dans une image.
"""
import math
import copy
import time
import numpy
from PIL import Image, ImageDraw
from .queue_binom import tabule_queue_binom
from .geometrie import Point
from .detection_segment_segangle import SegmentBord
from .detection_nfa import LigneGradient, InformationPoint


def convert_array2PIL(img, mode=None):
    """
    Convertit une image donnée sous la forme d'un array
    au format :epkg:`numpy:array`.

    @param      img     :epkg:`numpy:array`
    @param      mode    voir `modes <https://pillow.readthedocs.io/en/3.1.x/handbook/concepts.html#modes>`_,
                        si None, essaye de deviner.
    @return             *PIL*

    Le mode ``'binary'`` convertit une image issue
    de la fonction @see fn random_noise_image.
    """
    if mode == 'binary':
        fimg = img.astype(numpy.float32)
        img255 = (- fimg + 1) * 255
        img = img255.astype(numpy.uint8)
        mode = None
    return _load_image(img, 'PIL', mode=mode)


def convert_PIL2array(img):
    """
    Convertit une image donnée sous la forme d'une image :epkg:`Pillow`
    au format :epkg:`numpy:array`.

    @param      img     :epkg:`Pillow`
    @return             :epkg:`numpy:array`
    """
    return _load_image(img, 'array')


def _load_image(img, format='PIL', mode=None):
    """
    Charge une image en différents formats.

    @param      img     image (*array*, *PIL*, filename)
    @param      format  *array* ou *PIL*
    @param      mode    voir `modes <https://pillow.readthedocs.io/en/3.1.x/handbook/concepts.html#modes>`_,
                        si None, essaye de deviner.
    @return             *PIL*
    """
    if isinstance(img, str):
        img = Image.open(img)
        return _load_image(img, format)
    if isinstance(img, Image.Image):
        if format == 'PIL':
            return img
        if format == 'array':
            d1, d0 = img.size[1], img.size[0]
            img = numpy.array(img.getdata(), dtype=numpy.uint8)
            if len(img.shape) == 1:
                gray = img.shape[0] - d1 * d0
            elif len(img.shape) == 2:
                gray = img.shape[0] * img.shape[1] - d1 * d0
            elif(img.shape) == 3:
                gray = img.shape[0] * img.shape[1] * img.shape[2] - d1 * d0
            else:
                raise ValueError(  # pragma: no cover
                    "Unexpected shape {0}".format(img.shape))
            if gray == 0:
                img = img.reshape((d1, d0))
            else:
                img = img.reshape((d1, d0, 3))
            return img
        raise ValueError(  # pragma: no cover
            "Unexpected value for fomat: '{0}'".format(format))
    if isinstance(img, numpy.ndarray):
        if format == 'array':
            return img
        if format == 'PIL':
            return Image.fromarray(img, mode=mode)
        raise ValueError(  # pragma: no cover
            "Unexpected value for fomat: '{0}'".format(format))
    raise TypeError(  # pragma: no cover
        "numpy array expected not {0}".format(type(img)))


def compute_gradient(img, color=None):
    """
    Retourne le gradient d'une image sous forme d'une matrice
    de Point, consideres ici comme des vecteurs.
    """
    return _calcule_gradient(img, color=color)


def _calcule_gradient(img, color=None):
    """
    Retourne le gradient d'une image sous forme d'une matrice
    de Point, consideres ici comme des vecteurs.

    @param      img     *fichier*, *array*, *PIL* (image en niveau de gris)
    @param      method  ``'fast'`` or not
    @param      color   calcule le gradient pour cette couleur, None
                        si l'image est en niveau de gris
    @return             array of *shape (y, x, 2)*, first dimension is *dx*,
                        second one is *dy*
    """
    img = _load_image(img, 'array')
    img = img.astype(numpy.float32)
    if color is not None:
        img = img[:, :, color]

    dx1 = img[:, 1:-1] - img[:, :-2]
    dx2 = img[:, 2:] - img[:, 1:-1]
    dx = (dx1 + dx2) / 2

    dy1 = img[1:-1, :] - img[:-2, :]
    dy2 = img[2:, :] - img[1:-1, :]
    dy = (dy1 + dy2) / 2
    res = numpy.zeros(img.shape + (2,))
    res[:, 1:-1, 0] = dx
    res[1:-1, :, 1] = dy
    return res


def plot_gradient(image, gradient, more=None, direction=-1):
    """
    Construit une image a partir de la matrice de gradient
    afin de pouvoir l'afficher grace au module pygame,
    cette fonction place directement le resultat dans image,
    si direction > 0, cette fonction affiche egalement le gradient sur
    l'image tous les 10 pixels si direction vaut 10.
    """
    image_ = _load_image(image, 'PIL')
    image = ImageDraw.Draw(image_)
    X, Y = image_.size
    if direction != -1:
        for x in range(0, X - 1):
            for y in range(0, Y - 1):
                n = gradient[y, x]
                if more is None:
                    v = int((n[0]**2 + n[1] ** 2)**0.5 + 0.5)
                elif more == "x":
                    v = int(n[0] / 2 + 127 + 0.5)
                else:
                    v = int(n[1] / 2 + 127 + 0.5)
                image.line([(x, y), (x, y)], fill=(v, v, v))
    if direction in (0, -1):
        pass
    elif direction > 0:
        # on dessine des petits gradients dans l'image
        for x in range(0, X, direction):
            for y in range(0, Y, direction):
                n = gradient[y, x]
                t = (n[0]**2 + n[1] ** 2)**0.5
                if t == 0:
                    continue
                m = copy.copy(n)
                m /= t
                if t > direction:
                    t = direction
                if t < 2:
                    t = 2
                m *= t
                image.line([(x, y), (x + int(m[0]), y + int(m[1]))],
                           fill=(255, 255, 0))
    elif direction == -2:
        # derniere solution, la couleur represente l'orientation
        # en chaque point de l'image
        for x in range(0, X):
            for y in range(0, Y):
                n = gradient[y, x]
                i = int(-n[0] * 10 + 128)
                j = int(n[1] * 10 + 128)
                i, j = min(i, 255), min(j, 255)
                i, j = max(i, 0), max(j, 0)
                image.line([(x, y), (x, y)], fill=(0, j, i))
    else:
        raise ValueError(  # pragma: no cover
            "Unexpected value for direction={0}".format(direction))

    return image_


def plot_segments(image, segments, outfile=None, color=(255, 0, 0)):
    """
    Dessine les segments produits par la fonction
    @see fn detect_segments

    @param  image       image (*fichier*, *array*, *PIL*)
    @param  segments    résultats de la fonction @see fn detect_segments
    @param  outfile     fichier de sortie
    @param  color       couleur
    @return             nom de fichier ou image
    """
    image = _load_image(image, 'PIL')
    draw = ImageDraw.Draw(image)
    for seg in segments:
        draw.line([(seg.a.x, seg.a.y), (seg.b.x, seg.b.y)], fill=color)
    if outfile is not None:
        image.save(outfile)
        return outfile
    return image


def detect_segments(image, proba_bin=1.0 / 16,
                    cos_angle=math.cos(1.0 / 16 / 2 * (math.pi * 2)),
                    seuil_nfa=1e-5, seuil_norme=2, angle=math.pi / 24.0,
                    stop=-1, verbose=False):
    """
    Détecte les segments dans une image.

    @param  image       image (*fichier*, *array*, *PIL*)
    @param  proba_bin   est en fait un secteur angulaire (360 / 16)
                        qui determine la proximite de deux directions
    @param  cos_angle   est le cosinus de l'angle correspondant à ce secteur angulaire
    @param  seuil_nfa   au delà de ce seuil, on considere qu'un segment
                        génère trop de fausses alertes pour être sélectionné
    @param  seuil_norme norme en deça de laquelle un gradient est trop
                        petit pour etre significatif (c'est du bruit)
    @param  angle       lorsqu'on balaye l'image pour détecter les segments,
                        on tourne en rond selon les angles 0, angle, 2*angle,
                        3*angle, ...
    @param  stop        arrête après avoir collecté tant de segments
    @param  verbose     affiche l'avancement
    @return             les segments
    """
    gray_image = _load_image(image, 'PIL').convert('L')
    grad = _calcule_gradient(gray_image)

    # on calcule les tables de la binomiale pour eviter d'avoir a le fait a
    # chaque fois qu'on en a besoin
    yy, xx = grad.shape[:2]
    nbbin = int(math.ceil(math.sqrt(xx * xx + yy * yy)))
    binomiale = tabule_queue_binom(nbbin, proba_bin)

    # nb_seg est le nombre total de segment de l'image
    # il y a xx * yy pixels possibles dont (xx*yy)^2 couples de pixels (donc de segments)
    nb_seg = xx * xx * yy * yy

    # on cree une instance de la classe permettant de parcourir
    # tous les segments de l'image reliant deux points du contour
    seg = SegmentBord(Point(xx, yy))

    # initialisation avant de parcourir l'image
    segment = []  # resultat, ensemble des segments significatifs
    ti = time.perf_counter()  # memorise l'heure de depart
    # pour savoir combien de segments on a deja visite (seg)
    n = 0
    cont = True         # condition d'arret de la boucle

    # on cree une classe permettant de recevoir les informations relatives
    # a l'image et au gradient pour un segment reliant deux points
    # du contour de l'image
    points = [InformationPoint(Point(0, 0), False, 0)
              for i in range(0, xx + yy)]
    ligne = LigneGradient(points, seuil_norme=seuil_norme, seuil_nfa=seuil_nfa)

    # premier segment
    seg.premier()

    # autres variables a decouvrir en cours de route
    not_aligned = 0

    # tant qu'on a pas fini
    while cont:

        # calcule les informations relative a un segment de l'image reliant deux bords
        # position des pixels, norme du gradient, alignement avec le segment
        seg.decoupe_gradient(grad, cos_angle, ligne, seuil_norme)

        if len(ligne) > 3 and ligne.has_aligned_point():
            # si le segment contient plus de trois pixels
            # alors on peut se demander s'il inclut des sous-segments significatifs
            res = ligne.segments_significatifs(binomiale, nb_seg)

            # on ajoute les resultats à la liste
            segment.extend(res)
            if len(segment) >= stop > 0:
                break
        else:
            not_aligned += 1

        # on passe au segment suivant
        cont = seg.next()  # pylint: disable=E1102
        n += 1

        # pour verifier que cela avance
        if verbose and n % 1000 == 0:
            print(  # pragma: no cover
                "n = ", n, " ... ", len(segment), " temps ",
                "%2.2f" % (time.perf_counter() - ti), " sec",
                "nalign", not_aligned)

    return segment



# -*- coding: utf-8 -*-
"""
@file
@brief Conversion from tree to neural network.
"""
import numpy
import numpy.random as rnd
from scipy.special import expit, softmax, kl_div as kl_fct  # pylint: disable=E0611
from ._neural_tree_api import _TrainingAPI


class NeuralTreeNode(_TrainingAPI):
    """
    One node in a neural network.
    """

    @staticmethod
    def _relu(x):
        "Relu function."
        return numpy.maximum(x, 0)

    @staticmethod
    def _leakyrelu(x):
        "Leaky Relu function."
        return numpy.maximum(x, 0) + numpy.minimum(x, 0) * 0.01

    @staticmethod
    def _drelu(x):
        "Derivative of the Relu function."
        res = numpy.ones(x.shape, dtype=x.dtype)
        res[x < 0] = 0.
        return res

    @staticmethod
    def _dleakyrelu(x):
        "Derivative of the Leaky Relu function."
        res = numpy.ones(x.shape, dtype=x.dtype)
        res[x < 0] = 0.01
        return res

    @staticmethod
    def _dsigmoid(x):
        "Derivativ of the sigmoid function."
        y = expit(x)
        return y * (1 - y)

    @staticmethod
    def _softmax(x):
        "Derivative of the softmax function."
        if len(x.shape) == 2:
            return softmax(x, axis=1)
        return softmax(x)

    @staticmethod
    def _dsoftmax(x):
        "Derivative of the softmax function."
        soft = softmax(x)
        grad = - soft @ soft.T
        diag = numpy.diag(soft)
        return diag + grad

    @staticmethod
    def get_activation_function(activation):
        """
        Returns the activation function.
        It returns a function *y=f(x)*.
        """
        if activation == 'softmax':
            return NeuralTreeNode._softmax
        if activation == 'softmax4':
            return lambda x: NeuralTreeNode._softmax(x * 4)
        if activation in {'logistic', 'expit', 'sigmoid'}:
            return expit
        if activation == 'sigmoid4':
            return lambda x: expit(x * 4)
        if activation == 'relu':
            return NeuralTreeNode._relu
        if activation == 'leakyrelu':
            return NeuralTreeNode._leakyrelu
        if activation == 'identity':
            return lambda x: x
        raise ValueError(  # pragma: no cover
            "Unknown activation function '{}'.".format(activation))

    @staticmethod
    def get_activation_gradient_function(activation):
        """
        Returns the activation function.
        It returns a function *y=f'(x)*.
        About the sigmoid:

        .. math::

            \\begin{array}{l}
            f(x) &=& \frac{1}{1 + e^{-x}} \\\\
            f'(x) &=& \frac{e^{-x}}{(1 + e^{-x})^2} = f(x)(1-f(x))
            \\end{array}}
        """
        if activation == 'softmax':
            return NeuralTreeNode._dsoftmax
        if activation == 'softmax4':
            return lambda x: NeuralTreeNode._dsoftmax(x) * 4
        if activation in {'logistic', 'expit', 'sigmoid'}:
            return NeuralTreeNode._dsigmoid
        if activation == 'sigmoid4':
            return lambda x: NeuralTreeNode._dsigmoid(x) * 4
        if activation == 'relu':
            return NeuralTreeNode._drelu
        if activation == 'leakyrelu':
            return NeuralTreeNode._dleakyrelu
        if activation == 'identity':
            return lambda x: numpy.ones(x.shape, dtype=x.dtype)
        raise ValueError(  # pragma: no cover
            "Unknown activation gradient function '{}'.".format(activation))

    @staticmethod
    def get_activation_loss_function(activation):
        """
        Returns a default loss function based on the activation
        function. It returns two functions *g=loss(x,y)*.
        """
        if activation in {'logistic', 'expit', 'sigmoid', 'sigmoid4'}:
            # regression + regularization
            return lambda x, y: (x - y) ** 2
        if activation in {'softmax', 'softmax4'}:
            cst = numpy.finfo(numpy.float32).eps

            # classification
            def kl_fct2(x, y):
                return kl_fct(x + cst, y + cst)
            return kl_fct2
        if activation in {'identity', 'relu', 'leakyrelu'}:
            # regression
            return lambda x, y: (x - y) ** 2
        raise ValueError(
            "Unknown activation function '{}'.".format(activation))

    @staticmethod
    def get_activation_dloss_function(activation):
        """
        Returns the derivative of the default loss function based
        on the activation function. It returns a function
        *df(x,y)/dw, df(w)/dw* where *w* are the weights.
        """
        if activation in {'logistic', 'expit', 'sigmoid', 'sigmoid4'}:
            # regression + regularization
            def dregrdx(x, y):
                return (x - y) * 2

            return dregrdx

        if activation in {'softmax', 'softmax4'}:
            # classification
            cst = numpy.finfo(numpy.float32).eps

            def dclsdx(x, y):
                return numpy.log(x + cst) - numpy.log(y + cst)

            return dclsdx

        if activation in {'identity', 'relu', 'leakyrelu'}:
            # regression
            def dregdx(x, y):
                return (x - y) * 2

            return dregdx
        raise ValueError(  # pragma: no cover
            "Unknown activation function '{}'.".format(activation))

    def __init__(self, weights, bias=None, activation='sigmoid', nodeid=-1,
                 tag=None):
        """
        @param      weights     weights
        @param      bias        bias, if None, draws a random number
        @param      activation  activation function
        @param      nodeid      node id
        @param      tag         unused but to add information
                                on how this node was created
        """
        self.tag = tag
        if isinstance(weights, int):
            if activation.startswith('softmax'):
                weights = rnd.randn(2, weights)
            else:
                weights = rnd.randn(weights)
        if isinstance(weights, list):
            weights = numpy.array(weights)

        if len(weights.shape) == 1:
            self.n_outputs = 1
            if bias is None:
                bias = rnd.randn()
            self.coef = numpy.empty(len(weights) + 1)
            self.coef[1:] = weights
            self.coef[0] = bias

        elif len(weights.shape) == 2:
            self.n_outputs = weights.shape[0]
            if self.n_outputs == 1:
                raise RuntimeError(  # pragma: no cover
                    "Unexpected unsqueezed weights shape: {}".format(weights.shape))
            if bias is None:
                bias = rnd.randn(self.n_outputs)
            shape = list(weights.shape)
            shape[1] += 1
            self.coef = numpy.empty(shape)
            self.coef[:, 1:] = weights
            self.coef[:, 0] = bias
        else:
            raise RuntimeError(  # pragma: no cover
                "Unexpected weights shape: {}".format(weights.shape))

        self.activation = activation
        self.nodeid = nodeid
        self._set_fcts()

    def _set_fcts(self):
        self.activation_ = NeuralTreeNode.get_activation_function(
            self.activation)
        self.gradient_ = NeuralTreeNode.get_activation_gradient_function(
            self.activation)
        self.losss_ = NeuralTreeNode.get_activation_loss_function(
            self.activation)
        self.dlossds_ = NeuralTreeNode.get_activation_dloss_function(
            self.activation)

    @property
    def input_weights(self):
        "Returns the weights."
        if self.n_outputs == 1:
            return self.coef[1:]
        return self.coef[:, 1:]

    @property
    def bias(self):
        "Returns the weights."
        if self.n_outputs == 1:
            return self.coef[0]
        return self.coef[:, 0]

    def __getstate__(self):
        "usual"
        return {
            'coef': self.coef, 'activation': self.activation,
            'nodeid': self.nodeid, 'n_outputs': self.n_outputs,
            'tag': self.tag}

    def __setstate__(self, state):
        "usual"
        self.coef = state['coef']
        self.activation = state['activation']
        self.nodeid = state['nodeid']
        self.n_outputs = state['n_outputs']
        self.tag = state['tag']
        self._set_fcts()

    def __eq__(self, obj):
        if self.coef.shape != obj.coef.shape:
            return False
        if any(map(lambda xy: xy[0] != xy[1],
                   zip(self.coef.ravel(), obj.coef.ravel()))):
            return False
        if self.activation != obj.activation:
            return False
        return True

    def __repr__(self):
        "usual"
        if len(self.coef.shape) == 1:
            return "%s(weights=%r, bias=%r, activation=%r)" % (
                self.__class__.__name__, self.coef[1:],
                self.coef[0], self.activation)
        return "%s(weights=%r, bias=%r, activation=%r)" % (
            self.__class__.__name__, self.coef[:, 1:],
            self.coef[:, 0], self.activation)

    def _predict(self, X):
        "Computes inputs of the activation function."
        if self.n_outputs == 1:
            return X @ self.coef[1:] + self.coef[0]
        return (X.reshape((1, -1)) @ self.coef[:, 1:].T + self.coef[:, 0]).ravel()

    def predict(self, X):
        "Computes neuron outputs."
        if self.n_outputs == 1:
            return self.activation_(X @ self.coef[1:] + self.coef[0])
        if len(X.shape) == 2:
            return self.activation_(
                (X @ self.coef[:, 1:].T + self.coef[:, 0]))
        return self.activation_(
            (X.reshape((1, -1)) @ self.coef[:, 1:].T + self.coef[:, 0]).ravel())

    @property
    def ndim(self):
        "Returns the input dimension."
        if len(self.coef.shape) == 1:
            return self.coef.shape[0] - 1
        return self.coef.shape[1] - 1

    @property
    def training_weights(self):
        "Returns the weights stored in the neuron."
        return self.coef.ravel()

    def update_training_weights(self, X, add=True):  # pylint: disable=W0237
        """
        Updates weights.

        :param grad: vector to add to the weights such as gradient
        :param add: addition or replace
        """
        if add:
            self.coef += X.reshape(self.coef.shape)
        else:
            numpy.copyto(self.coef, X.reshape(self.coef.shape))

    def fill_cache(self, X):
        """
        Creates a cache with intermediate results.
        ``lX`` is the results before the activation function,
        ``aX`` is the results after the activation function, the prediction.
        """
        cache = dict(lX=self._predict(X))
        cache['aX'] = self.activation_(cache['lX'])
        return cache

    def _common_loss_dloss(self, X, y, cache=None):
        """
        Common beginning to methods *loss*, *dlossds*,
        *dlossdw*.
        """
        if cache is not None and 'aX' in cache:
            act = cache['aX']
        else:
            act = self.predict(X)
        return act

    def loss(self, X, y, cache=None):
        """
        Computes the loss. Returns a float.
        """
        act = self._common_loss_dloss(X, y, cache=cache)
        if len(X.shape) == 1:
            return self.losss_(act, y)  # pylint: disable=E1120
        return self.losss_(act, y)  # pylint: disable=E1120

    def dlossds(self, X, y, cache=None):
        """
        Computes the loss derivative due to prediction error.
        """
        act = self._common_loss_dloss(X, y, cache=cache)
        return self.dlossds_(act, y)

    def gradient_backward(self, graddx, X, inputs=False, cache=None):
        """
        Computes the gradients at point *X*.

        :param graddx: existing gradient against the inputs
        :param X: computes the gradient in X
        :param inputs: if False, derivative against the coefficients,
            otherwise against the inputs.
        :param cache: cache intermediate results
        :return: gradient
        """
        if cache is None:
            cache = self.fill_cache(X)

        pred = cache['aX']
        ga = self.gradient_(pred)
        if len(ga.shape) == 2:
            f = graddx @ ga
        else:
            f = graddx * ga

        if inputs:
            if len(self.coef.shape) == 1:
                rgrad = numpy.empty(X.shape)
                rgrad[:] = self.coef[1:]
                rgrad *= f
            else:
                rgrad = numpy.sum(
                    self.coef[:, 1:] * f.reshape((-1, 1)), axis=0)
            return rgrad

        rgrad = numpy.empty(self.coef.shape)
        if len(self.coef.shape) == 1:
            rgrad[0] = 1
            rgrad[1:] = X
            rgrad *= f
        else:
            rgrad[:, 0] = 1
            rgrad[:, 1:] = X
            rgrad *= f.reshape((-1, 1))
        return rgrad



# -*- coding: utf-8 -*-
"""
@file
@brief Ce module construit les probabilités d'une loi binomiale :math:`B(n,p)`.
"""


def tabule_queue_binom(n, p):
    """
    Retourne un dictionnaire dont la clé est couple d'entiers *(a,b)*
    si *t* est le resultat, alors :math:`t=[(a,b)]` est la probabilité
    qu'il y ait *b* événements parmi *a* sachant que la probabilité d'un
    événement est *p* : :math:`t [ (a,b) ] = C_a^b p^b (1-p)^ {(a-b)}`

    Pour aller plus vite, ces probabilités sont estimées par récurrence :

    * :math:`\\forall m, \\; t [(m,0)] = 1.0`
    * :math:`\\forall m, \\; t [(m,m+1)] = 0.0`
      et :math:`t[(m,k)] = p * t [ (m-1, k-1)] + (1-p) * t [ (m-1,k) ]`

    Cette fonction calcule tous les coefficients :math:`t [ (a,b) ]` pour une
    probabilité :math:`p` donnée et :math:`b \\infegal a \\infegal n`.

    Ces probabilités sont stockées dans un dictionnaire car s'ils étaient
    stockées dans une matrice, celle-ci serait triangulaire inférieure.
    """
    t = {}
    t[(0, 0)] = 1.0
    t[(0, 1)] = 0.0
    for m in range(1, n + 1):
        t[(m, 0)] = 1.0
        t[(m, m + 1)] = 0.0
        for k in range(1, m + 1):
            t[(m, k)] = p * t[(m - 1, k - 1)] + (1 - p) * t[(m - 1, k)]
    return t



# -*- coding: utf-8 -*-
"""
@file
@brief About :epkg:`ROC`.
"""
import math
import itertools
from enum import Enum
import pandas
import numpy


class ROC:
    """
    Helper to draw a :epkg:`ROC` curve.
    """

    class CurveType(Enum):
        """
        Curve types:

        * *PROBSCORE*: 1 - False Positive / True Positive
        * *ERRPREC*: error / recall
        * *RECPREC*: precision / recall
        * *ROC*: False Positive / True Positive
        * *SKROC*: False Positive / True Positive
          (`scikit-learn
          <http://scikit-learn.org/stable/modules/generated/sklearn.metrics.roc_curve.html>`_)
        """
        PROBSCORE = 2
        ERRREC = 3
        RECPREC = 4
        ROC = 5
        SKROC = 6

    def __init__(self, y_true=None, y_score=None, sample_weight=None, df=None):
        """
        Initialisation with a dataframe and two or three columns:

        * column 1: score (y_score)
        * column 2: expected answer (boolean) (y_true)
        * column 3: weight (optional) (sample_weight)

        @param  y_true          if *df* is None, *y_true*, *y_score*, *sample_weight* must be filled,
                                *y_true* is whether or None the answer is true.
                                *y_true* means the prediction is right.
        @param  y_score         score prediction
        @param  sample_weight   weights
        @param  df              dataframe or array or list,
                                it must contains 2 or 3 columns always in the same order
        """
        if df is None:
            df = pandas.DataFrame()
            df["score"] = y_score
            df["label"] = y_true
            if sample_weight is not None:
                df["weight"] = sample_weight
            self.data = df
        elif isinstance(df, list):
            if len(df[0]) == 2:
                self.data = pandas.DataFrame(df, columns=["score", "label"])
            else:
                self.data = pandas.DataFrame(
                    df, columns=["score", "label", "weight"])
        elif isinstance(df, numpy.ndarray):
            if df.shape[1] == 2:
                self.data = pandas.DataFrame(df, columns=["score", "label"])
            else:
                self.data = pandas.DataFrame(
                    df, columns=["score", "label", "weight"])
        elif not isinstance(df, pandas.DataFrame):
            raise TypeError(  # pragma: no cover
                "df should be a DataFrame, not {0}".format(type(df)))
        else:
            self.data = df.copy()
        self.data.sort_values(self.data.columns[0], inplace=True)
        self.data.reset_index(drop=True, inplace=True)
        if self.data.shape[1] == 2:
            self.data["weight"] = 1.0

    @property
    def Data(self):
        """
        Returns the underlying dataframe.
        """
        return self.data

    def __len__(self):
        """
        usual
        """
        return len(self.data)

    def __repr__(self):
        """
        Shows first elements, precision rate.
        """
        return self.__str__()

    def __str__(self):
        """
        Shows first elements, precision rate.
        """
        rows = []
        rows.append("Overall precision: %3.2f - AUC=%f" %
                    (self.precision(), self.auc()))
        rows.append("--------------")
        rows.append(str(self.data.head(min(5, len(self)))))
        rows.append("--------------")
        rows.append(str(self.data.tail(min(5, len(self)))))
        rows.append("--------------")
        roc = self.compute_roc_curve(10, ROC.CurveType.ROC)
        rows.append(str(roc))
        rows.append("--------------")
        roc = self.compute_roc_curve(10, ROC.CurveType.ERRREC)
        rows.append(str(roc))
        return "\n".join(rows)

    def confusion(self, score=None, nb=10, curve=CurveType.ROC, bootstrap=False):
        """
        Computes the confusion matrix for a specific *score*
        or all if *score* is None.

        @param      score       score or None.
        @param      nb          number of scores (if *score* is None)
        @param      curve       see :class:`CurveType <mlstatpy.ml.roc.ROC.CurveType>`
        @param      boostrap    builds the curve after resampling
        @return                 One row if score is precised, many roww is score is None
        """
        if not bootstrap:
            cloud = self.data.copy()
        else:
            cloud = self.random_cloud()

        if score is None:

            sum_weights = cloud[cloud.columns[2]].sum()
            if nb <= 0:
                nb = len(cloud)
            else:
                nb = min(nb, len(cloud))
            seuil = numpy.arange(nb + 1) * sum_weights / nb

            cloud = cloud.iloc[::-1].copy()
            cloud["lw"] = cloud[cloud.columns[2]] * cloud[cloud.columns[1]]
            cloud["cw"] = cloud[cloud.columns[2]].cumsum()
            cloud["clw"] = cloud["lw"].cumsum()
            if cloud.columns[4] != "cw":
                raise ValueError(
                    "Column 4 should be 'cw'.")  # pragma: no cover
            if cloud.columns[5] != "clw":
                raise ValueError(
                    "Column 5 should be 'clw'.")  # pragma: no cover

            pos_roc = 0
            pos_seuil = 0
            if curve is ROC.CurveType.ROC:
                roc = pandas.DataFrame(0, index=numpy.arange(
                    nb + 2), columns=["True Positive", "False Positive",
                                      "False Negative", "True Negative",
                                      "threshold"])
                sum_good_weights = cloud.iloc[-1, 5]
                sum_bad_weights = sum_weights - sum_good_weights
                roc.iloc[0, 0] = 0
                roc.iloc[0, 1] = 0
                roc.iloc[0, 2] = sum_good_weights
                roc.iloc[0, 3] = sum_bad_weights
                roc.iloc[0, 4] = max(cloud.iloc[:, 0])
                pos_roc += 1
                for i in range(len(cloud)):
                    if cloud.iloc[i, 4] > seuil[pos_seuil]:
                        tp = cloud.iloc[i, 5]
                        fp = cloud.iloc[i, 4] - cloud.iloc[i, 5]
                        roc.iloc[pos_roc, 0] = tp
                        roc.iloc[pos_roc, 1] = fp
                        roc.iloc[pos_roc, 2] = sum_good_weights - tp
                        roc.iloc[pos_roc, 3] = sum_bad_weights - fp
                        roc.iloc[pos_roc, 4] = cloud.iloc[i, 0]
                        pos_roc += 1
                        pos_seuil += 1
                roc.iloc[pos_roc:, 0] = sum_good_weights
                roc.iloc[pos_roc:, 1] = sum_bad_weights
                roc.iloc[pos_roc:, 2] = 0
                roc.iloc[pos_roc:, 3] = 0
                roc.iloc[pos_roc:, 4] = min(cloud.iloc[:, 0])
                return roc
            raise NotImplementedError(  # pragma: no cover
                "Unexpected type '{0}', only ROC is allowed.".format(curve))

        # if score is not None
        roc = self.confusion(nb=len(self), curve=curve,
                             bootstrap=False, score=None)
        roc = roc[roc["threshold"] <= score]
        if len(roc) == 0:
            raise ValueError(  # pragma: no cover
                "The requested confusion is empty for score={0}.".format(score))
        return roc[:1]

    def precision(self):
        """
        Computes the precision.
        """
        score, weight = self.data.columns[0], self.data.columns[2]
        return (self.data[score] * self.data[weight] * 1.0).sum() / self.data[weight].sum()

    def compute_roc_curve(self, nb=100, curve=CurveType.ROC, bootstrap=False):
        """
        Computes a ROC curve with *nb* points avec nb,
        if *nb == -1*, there are as many as points as the data contains,
        if *bootstrap == True*, it draws random number to create confidence
        interval based on bootstrap method.

        @param      nb          number of points for the curve
        @param      curve       see :class:`CurveType <mlstatpy.ml.roc.ROC.CurveType>`
        @param      boostrap    builds the curve after resampling
        @return                 DataFrame (metrics and threshold)

        If *curve* is *SKROC*, the parameter *nb* is not taken into account.
        It should be set to 0.
        """
        if curve is ROC.CurveType.ERRREC:
            roc = self.compute_roc_curve(
                nb=nb, curve=ROC.CurveType.RECPREC, bootstrap=bootstrap)
            roc["error"] = - roc["precision"] + 1
            return roc[["error", "recall", "threshold"]]
        if curve is ROC.CurveType.PROBSCORE:
            roc = self.compute_roc_curve(
                nb=nb, curve=ROC.CurveType.ROC, bootstrap=bootstrap)
            roc["P(->s)"] = roc["False Positive Rate"]
            roc["P(+<s)"] = - roc["True Positive Rate"] + 1
            return roc[["P(+<s)", "P(->s)", "threshold"]]

        if not bootstrap:
            cloud = self.data.copy()
        else:
            cloud = self.random_cloud()

        if curve is ROC.CurveType.SKROC:
            if nb > 0:
                raise NotImplementedError(  # pragma: no cover
                    "nb must be <= 0 si curve is SKROC")
            from sklearn.metrics import roc_curve
            fpr, tpr, thresholds = roc_curve(y_true=cloud[cloud.columns[1]],
                                             y_score=cloud[cloud.columns[0]],
                                             sample_weight=cloud[cloud.columns[2]])
            roc = pandas.DataFrame(0, index=numpy.arange(len(fpr)),
                                   columns=["False Positive Rate", "True Positive Rate", "threshold"])
            roc_cols = list(roc.columns)
            roc[roc_cols[0]] = fpr
            roc[roc_cols[1]] = tpr
            roc[roc_cols[2]] = thresholds
            return roc

        sum_weights = cloud[cloud.columns[2]].sum()
        if nb <= 0:
            nb = len(cloud)
        else:
            nb = min(nb, len(cloud))
        seuil = numpy.arange(nb + 1) * sum_weights / nb

        cloud = cloud.iloc[::-1].copy()
        cloud["lw"] = cloud[cloud.columns[2]] * cloud[cloud.columns[1]]
        cloud["cw"] = cloud[cloud.columns[2]].cumsum()
        cloud["clw"] = cloud["lw"].cumsum()
        sum_weights_ans = cloud["lw"].sum()
        if cloud.columns[4] != "cw":
            raise ValueError("Column 4 should be 'cw'.")  # pragma: no cover
        if cloud.columns[5] != "clw":
            raise ValueError("Column 5 should be 'clw'.")  # pragma: no cover

        pos_roc = 0
        pos_seuil = 0

        if curve is ROC.CurveType.ROC:
            roc = pandas.DataFrame(0, index=numpy.arange(
                nb + 1), columns=["False Positive Rate", "True Positive Rate", "threshold"])
            sum_good_weights = cloud.iloc[-1, 5]
            sum_bad_weights = sum_weights - sum_good_weights
            for i in range(len(cloud)):
                if cloud.iloc[i, 4] > seuil[pos_seuil]:
                    ds = cloud.iloc[i, 4] - cloud.iloc[i, 5]
                    roc.iloc[pos_roc, 0] = ds / sum_bad_weights
                    roc.iloc[pos_roc, 1] = cloud.iloc[i, 5] / sum_good_weights
                    roc.iloc[pos_roc, 2] = cloud.iloc[i, 0]
                    pos_roc += 1
                    pos_seuil += 1
            roc.iloc[pos_roc:, 0] = (
                cloud.iloc[-1, 4] - cloud.iloc[-1, 5]) / sum_bad_weights
            roc.iloc[pos_roc:, 1] = cloud.iloc[-1, 5] / sum_good_weights
            roc.iloc[pos_roc:, 2] = cloud.iloc[-1, 0]

        elif curve is ROC.CurveType.RECPREC:
            roc = pandas.DataFrame(0, index=numpy.arange(
                nb + 1), columns=["recall", "precision", "threshold"])
            for i in range(len(cloud)):
                if cloud.iloc[i, 4] > seuil[pos_seuil]:
                    roc.iloc[pos_roc, 0] = cloud.iloc[i, 4] / sum_weights
                    if cloud.iloc[i, 4] > 0:
                        roc.iloc[pos_roc, 1] = cloud.iloc[
                            i, 5] / cloud.iloc[i, 4]
                    else:
                        roc.iloc[pos_roc, 1] = 0.0
                    roc.iloc[pos_roc, 2] = cloud.iloc[i, 0]
                    pos_roc += 1
                    pos_seuil += 1
            roc.iloc[pos_roc:, 0] = 1.0
            roc.iloc[pos_roc:, 1] = sum_weights_ans / sum_weights
            roc.iloc[pos_roc:, 2] = cloud.iloc[-1, 0]

        else:
            raise NotImplementedError(  # pragma: no cover
                "Unknown curve type '{}'.".format(curve))

        return roc

    def random_cloud(self):
        """
        Resamples among the data.

        @return      DataFrame
        """
        res = self.data.sample(len(self.data), weights=self.data[
                               self.data.columns[2]], replace=True)
        return res.sort_values(res.columns[0])

    def plot(self, nb=100, curve=CurveType.ROC, bootstrap=0,
             ax=None, thresholds=False, **kwargs):
        """
        Plots a :epkg:`ROC` curve.

        @param      nb          number of points
        @param      curve       see :class:`CurveType <mlstatpy.ml.roc.ROC.CurveType>`
        @param      boostrap    number of curves for the boostrap (0 for None)
        @param      ax          axis
        @param      thresholds  use thresholds for the X axis
        @param      kwargs      sent to `pandas.plot <http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.plot.html>`_
        @return                 ax
        """
        nb_bootstrap = 0
        if bootstrap > 0:
            ckwargs = kwargs.copy()
            if 'color' not in ckwargs:
                ckwargs['color'] = 'r'
            if 'linewidth' not in kwargs:
                ckwargs['linewidth'] = 0.2
            ckwargs['legend'] = False
            if 'label' in ckwargs:
                del ckwargs['label']
            for _ in range(0, bootstrap):
                roc = self.compute_roc_curve(nb, curve=curve, bootstrap=True)
                if thresholds:
                    cols = list(_ for _ in roc.columns if _ != "threshold")
                    roc = roc.sort_values("threshold").reset_index(drop=True)
                    ax = roc.plot(x="threshold", y=cols,
                                  ax=ax, label=['_nolegend_' for i in cols], **ckwargs)
                else:
                    cols = list(_ for _ in roc.columns[
                                1:] if _ != "threshold")
                    roc = roc.sort_values(
                        roc.columns[0]).reset_index(drop=True)
                    ax = roc.plot(x=roc.columns[0], y=cols,
                                  ax=ax, label=['_nolegend_' for i in cols], **ckwargs)
                nb_bootstrap += len(cols)
            bootstrap = 0

        if bootstrap <= 0:
            if 'legend' not in kwargs:
                kwargs['legend'] = False
            roc = self.compute_roc_curve(nb, curve=curve)
            if not thresholds:
                roc = roc[[_ for _ in roc.columns if _ != "threshold"]]

            cols = list(_ for _ in roc.columns if _ != "threshold")
            final = 0
            if thresholds:
                if 'label' in kwargs and len(cols) != len(kwargs['label']):
                    raise ValueError(  # pragma: no cover
                        'label must have {0} values'.format(len(cols)))
                roc = roc.sort_values("threshold").reset_index(drop=True)
                ax = roc.plot(x="threshold", y=cols, ax=ax, **kwargs)
                ax.set_ylim([0, 1])
                ax.set_xlabel("thresholds")
                final += len(cols)
                diag = 0
            else:
                if 'label' in kwargs and len(cols) - 1 != len(kwargs['label']):
                    raise ValueError(
                        'label must have {0} values'.format(len(cols) - 1))
                final += len(cols) - 1
                roc = roc.sort_values(cols[0]).reset_index(drop=True)
                ax = roc.plot(x=cols[0], y=cols[1:], ax=ax, **kwargs)
                if curve is ROC.CurveType.ROC or curve is ROC.CurveType.SKROC:
                    ax.plot([0, 1], [0, 1], "k--", linewidth=1)
                    diag = 1
                else:
                    diag = 0
                ax.set_xlabel(roc.columns[0])
                if len(roc.columns) == 2:
                    ax.set_ylabel(roc.columns[1])
                ax.set_xlim([0, 1])
                ax.set_ylim([0, 1])

            # legend
            handles, labels = ax.get_legend_handles_labels()
            tot = final + nb_bootstrap
            diag = len(handles) - diag
            handles = handles[:-tot] + handles[-final:diag]
            new_labels = labels[:-tot] + labels[-final:diag]
            ax.legend(handles, new_labels)

        return ax

    def auc(self, cloud=None):
        """
        Computes the area under the curve (:epkg:`AUC`).

        @param      cloud       data or None to use ``self.data``, the function
                                assumes the data is sorted.
        @return                 AUC

        The first column is the label, the second one is the score,
        the third one is the weight.
        """
        if cloud is None:
            cloud = self.data
        good = cloud[cloud[cloud.columns[1]] == 1]
        wrong = cloud[cloud[cloud.columns[1]] == 0]
        auc = 0.0
        for a, b in itertools.product(good.itertuples(False), wrong.itertuples(False)):
            if a[0] > b[0]:
                auc += a[2] * b[2]
            elif a[0] >= b[0]:
                auc += a[2] * b[2] / 2
        if auc == 0 and good.shape[0] + wrong.shape[0] < self.data.shape[0]:
            raise ValueError(  # pragma: no cover
                "Label are not right, expect 0 and 1 not {0}".format(
                    set(cloud[cloud.columns[1]])))
        n = len(wrong) * len(good)
        if n > 0:
            auc /= float(n)
        return auc

    def auc_interval(self, bootstrap=10, alpha=0.95):
        """
        Determines a confidence interval for the :epkg:`AUC` with bootstrap.

        @param      bootstrap       number of random estimation
        @param      alpha           define the confidence interval
        @return                     dictionary of values
        """
        if bootstrap <= 1:
            raise ValueError(
                "Use auc instead, bootstrap < 2")  # pragma: no cover
        rate = []
        for _ in range(0, bootstrap):
            cloud = self.random_cloud()
            auc = self.auc(cloud)
            rate.append(auc)

        rate.sort()
        ra = self.auc(self.data)

        i1 = int(alpha * len(rate) / 2)
        i2 = max(i1, len(rate) - i1 - 1)
        med = rate[len(rate) // 2]
        moy = float(sum(rate)) / len(rate)
        var = 0
        for r in rate:
            var += r * r
        var = float(var) / len(rate)
        var = var - moy * moy
        return dict(auc=ra, interval=(rate[i1], rate[i2]),
                    min=rate[0], max=rate[len(rate) - 1],
                    mean=moy, var=math.sqrt(var), mediane=med)

    def roc_intersect(self, roc, x):
        """
        The :epkg:`ROC` curve is defined by a set of points.
        This function interpolates those points to determine
        *y* for any *x*.

        @param      roc     ROC curve
        @param      x       x
        @return             y
        """
        below = roc[roc[roc.columns[0]] <= x]
        i = len(below)
        if i == len(roc):
            return 0.0

        p2 = tuple(roc.iloc[i, :])
        if i - 1 <= 0:
            p1 = (1, 1)
        else:
            p1 = tuple(roc.iloc[i - 1, :])

        if p1[0] == p2[0]:
            return (p1[1] + p2[0]) / 2
        return (x - p1[0]) / (p2[0] - p1[0]) * (p2[1] - p1[1]) + p1[1]

    def roc_intersect_interval(self, x, nb, curve=CurveType.ROC, bootstrap=10, alpha=0.05):
        """
        Computes a confidence interval for the value returned by
        @see me roc_intersect.

        @param      roc         ROC curve
        @param      x           x
        @param      curve       see :class:`CurveType <mlstatpy.ml.roc.ROC.CurveType>`
        @return                 dictionary
        """

        rate = []
        for _ in range(0, bootstrap):
            roc = self.compute_roc_curve(nb, curve=curve, bootstrap=True)
            r = self.roc_intersect(roc, x)
            rate.append(r)

        rate.sort()

        roc = self.compute_roc_curve(nb, curve=curve)
        ra = self.roc_intersect(roc, x)

        i1 = int(alpha * len(rate) / 2)
        i2 = max(i1, len(rate) - i1 - 1)
        med = rate[len(rate) // 2]
        moy = float(sum(rate)) / len(rate)
        var = 0
        for r in rate:
            var += r * r
        var = float(var) / len(rate)
        var = var - moy * moy
        return dict(y=ra, interval=(rate[i1], rate[i2]),
                    min=rate[0], max=rate[len(rate) - 1],
                    mean=moy, var=math.sqrt(var), mediane=med)



# -*- coding: utf-8 -*-
import sys
import os
from setuptools import find_packages, setup
from pyquicksetup import read_version, read_readme, default_cmdclass

#########
# settings
#########

project_var_name = "mlstatpy"
versionPython = "%s.%s" % (sys.version_info.major, sys.version_info.minor)
path = "Lib/site-packages/" + project_var_name
readme = 'README.rst'
history = 'HISTORY.rst'
requirements = None

KEYWORDS = project_var_name + ', Xavier Dupré'
DESCRIPTION = """Lectures about machine learning, mathematics, statistics, programming."""
CLASSIFIERS = [
    'Programming Language :: Python :: 3',
    'Intended Audience :: Developers',
    'Topic :: Scientific/Engineering',
    'Topic :: Education',
    'License :: OSI Approved :: MIT License',
    'Development Status :: 5 - Production/Stable'
]

#######
# data
#######

packages = find_packages('src', exclude='src')
package_dir = {k: "src/" + k.replace(".", "/") for k in packages}
package_data = {}


setup(
    name=project_var_name,
    version=read_version(__file__, project_var_name, subfolder='src'),
    author='Xavier Dupré',
    author_email='xavier.dupre@gmail.com',
    license="MIT",
    url="http://www.xavierdupre.fr/app/mlstatpy/helpsphinx/index.html",
    download_url="https://github.com/sdpython/mlstatpy/",
    description=DESCRIPTION,
    long_description=read_readme(__file__),
    cmdclass=default_cmdclass(),
    keywords=KEYWORDS,
    classifiers=CLASSIFIERS,
    packages=packages,
    package_dir=package_dir,
    package_data=package_data,
    setup_requires=["pyquicksetup"],
    install_requires=['numpy>=1.16', 'scipy>=1.4',
                      'mlinsights>=0.2', 'cpyquickhelper>=0.2'],
)



# -*- coding: utf-8 -*-
"""
@file
@brief Functions for :ref:`l-exemple_optim_alea`.
"""
import math
import random


def factorielle(x):
    """
    Calcule :math:`x!` de façon récursive.
    """
    if x == 0:
        return 1
    else:
        return x * factorielle(x - 1)


def profit(N, X, p, q, s):
    """
    Calcule le profit.

    @param      N       nombre de poulets vendus
    @param      X       nombre de poulets achetés
    @param      p       prix d'achat
    @param      q       prix de vente
    @param      s       prix soldé
    @return             profit
    """
    if X <= N:
        return X * (q - p)
    else:
        return X * (s - p) + N * (q - s)


def proba_poisson(lx, i):
    """
    Calcule la probabilité :math:`\\pr{X=i}``
    lorsque :math:`X` suit une loi de Poisson de paramètre
    :math:`\\lambda`.
    """
    return math.exp(-lx) * (lx ** i) / factorielle(i)


def esperance(X, p, q, s, lx):
    """
    Espérance du profit en faisant varier
    le nombre de poulet vendus.

    @param      X       nombre de poulets achetés
    @param      p       prix d'achat
    @param      q       prix de vente
    @param      s       prix soldé
    @param      lx      paramètre :math:`\\lambda`
    @return             espérance du profit
    """
    res = 0.0
    for i in range(0, lx * 2):
        res += profit(float(i), X, p, q, s) * proba_poisson(lx, i)
    return res


def maximum(p, q, s, lx):
    """
    Calcule les espérances de profit pour différents nombres
    de poulets achetés.

    @param      p       prix d'achat
    @param      q       prix de vente
    @param      s       prix soldé
    @param      lx      paramètre :math:`\\lambda`
    @return             liste ``(X, profit)``
    """
    res = []
    for X in range(0, 2 * lx):
        r = esperance(X, p, q, s, lx)
        res.append((X, r))
    return res


def find_maximum(res):
    """
    Trouver le couple (nombre de poulets achetés, profit)
    lorsque le profit est maximum.

    @param      res         résultat de la fonction :func:`maximum <mlstatpy.garden.poulet.maximum>`
    @return                 ``(X, profit)`` maximum
    """
    m = (0, 0)
    for r in res:
        if r[1] > m[1]:
            m = r
    return m


def exponentielle(lx):
    """
    Simule une loi exponentielle de paramètre :math:`\\lambda`.
    """
    u = random.random()
    return - 1.0 / lx * math.log(1.0 - u)


def poisson(lx):
    """
    Simule une loi de Poisson de paramètre :math:`\\lambda`.
    """
    s = 0
    i = 0
    while s <= 1:
        s += exponentielle(lx)
        i += 1
    return i - 1


def poisson_melange(params, coef):
    """
    Simule une variable selon un mélange de loi de Poisson.

    @param      params      liste de paramètre :math:`\\lambda`
    @param      coef        ``coef[i]`` coefficient associé à la loi de paramètre ``params[i]``
    @return                 valeur simulée
    """
    s = 0
    for i, pa in enumerate(params):
        p = poisson(pa)
        s += p * coef[i]
    return s


def histogramme_poisson_melange(params, coef, n=100000):
    """
    Calcule un histogramme d'un mélange de loi de Poisson.

    @param      params      liste de paramètre :math:`\\lambda`
    @param      coef        ``coef[i]`` coefficient associé à la loi de paramètre ``params[i]``
    @return                 histogramme
    """
    h = [0.0 for i in range(0, 4 * max(params))]
    for i in range(0, n):
        x = poisson_melange(params, coef)
        if x < len(h):
            h[x] += 1
    s = sum(h)
    for i in range(0, len(h)):
        h[i] = float(h[i]) / s
    return h


def f_proba_poisson_melange():
    """
    Wraps function *proba_poisson_melange* to avoid
    global variable.
    """

    proba_poisson_melange_tableau = []

    def local_proba_poisson_melange(params, coef, i):
        """
        Calcule la probabilité :math:`\\pr{X=i}``
        lorsque :math:`X` suit un mélange de lois.

        @param      params      liste de paramètre :math:`\\lambda`
        @param      coef        ``coef[i]`` coefficient associé à la loi de paramètre ``params[i]``
        @return                 valeur
        """
        if len(proba_poisson_melange_tableau) == 0:
            proba_poisson_melange_tableau.extend(
                histogramme_poisson_melange(params, coef))
        if i >= len(proba_poisson_melange_tableau):
            return 0.0  # pragma: no cover
        return proba_poisson_melange_tableau[i]

    return local_proba_poisson_melange


proba_poisson_melange = f_proba_poisson_melange()



# -*- coding: utf-8 -*-
"""
@file
@brief Conversion from tree to neural network.
"""
from io import BytesIO
import pickle
import numpy
from ._neural_tree_api import _TrainingAPI
from ._neural_tree_node import NeuralTreeNode


def label_class_to_softmax_output(y_label):
    """
    Converts a binary class label into a matrix
    with two columns of probabilities.

    .. runpython::
        :showcode:

        import numpy
        from mlstatpy.ml.neural_tree import label_class_to_softmax_output

        y_label = numpy.array([0, 1, 0, 0])
        soft_y = label_class_to_softmax_output(y_label)
        print(soft_y)
    """
    if len(y_label.shape) != 1:
        raise ValueError(
            "y_label must be a vector but has shape {}.".format(y_label.shape))
    y = numpy.empty((y_label.shape[0], 2), dtype=numpy.float64)
    y[:, 0] = (y_label < 0.5).astype(numpy.float64)
    y[:, 1] = 1 - y[:, 0]
    return y


class NeuralTreeNet(_TrainingAPI):
    """
    Node ensemble.

    .. runpython::
        :showcode:

        import numpy
        from mlstatpy.ml.neural_tree import NeuralTreeNode, NeuralTreeNet

        w1 = numpy.array([-0.5, 0.8, -0.6])

        neu = NeuralTreeNode(w1[1:], bias=w1[0], activation='sigmoid')
        net = NeuralTreeNet(2, empty=True)
        net.append(neu, numpy.arange(2))

        ide = NeuralTreeNode(numpy.array([1.]),
                             bias=numpy.array([0.]),
                             activation='identity')

        net.append(ide, numpy.arange(2, 3))

        X = numpy.abs(numpy.random.randn(10, 2))
        pred = net.predict(X)
        print(pred)
    """

    def __init__(self, dim, empty=True):
        """
        @param      dim     space dimension
        @param      empty   empty network, other adds an identity node
        """
        self.dim = dim
        if empty:
            self.nodes = []
            self.nodes_attr = []
        else:
            self.nodes = [
                NeuralTreeNode(
                    numpy.ones((dim,), dtype=numpy.float64),
                    bias=numpy.float64(0.),
                    activation='identity', nodeid=0)]
            self.nodes_attr = [dict(inputs=numpy.arange(0, dim), output=dim,
                                    coef_size=self.nodes[0].coef.size,
                                    first_coef=0)]
        self._update_members()

    def copy(self):
        st = BytesIO()
        pickle.dump(self, st)
        cop = BytesIO(st.getvalue())
        return pickle.load(cop)

    def _update_members(self, node=None, attr=None):
        "Updates internal members."
        if node is None or attr is None:
            if len(self.nodes_attr) == 0:
                self.size_ = self.dim
            else:
                self.size_ = max(d['output'] for d in self.nodes_attr) + 1
            self.output_to_node_ = {}
            self.input_to_node_ = {}
            for node2, attr2 in zip(self.nodes, self.nodes_attr):
                if isinstance(attr2['output'], list):
                    for o in attr2['output']:
                        self.output_to_node_[o] = node2, attr2
                else:
                    self.output_to_node_[attr2['output']] = node2, attr2
                for i in attr2['inputs']:
                    self.input_to_node_[i] = node2, attr2
        else:
            if len(node.input_weights.shape) == 1:
                self.size_ += 1
            else:
                self.size_ += node.input_weights.shape[0]
            if isinstance(attr['output'], list):
                for o in attr['output']:
                    self.output_to_node_[o] = node, attr
            else:
                self.output_to_node_[attr['output']] = node, attr
            for i in attr['inputs']:
                self.input_to_node_[i] = node, attr

    def __repr__(self):
        "usual"
        return "%s(%d)" % (self.__class__.__name__, self.dim)

    def clear(self):
        "Clear all nodes"
        del self.nodes[:]
        del self.nodes_attr[:]
        self._update_members()

    def append(self, node, inputs):
        """
        Appends a node into the graph.

        @param      node        node to add
        @param      inputs      index of input nodes
        """
        if len(node.input_weights.shape) == 1:
            if node.input_weights.shape[0] != len(inputs):
                raise RuntimeError(
                    "Dimension mismatch between weights [{}] and inputs [{}].".format(
                        node.input_weights.shape[0], len(inputs)))
            node.nodeid = len(self.nodes)
            self.nodes.append(node)
            first_coef = (
                0 if len(self.nodes_attr) == 0 else
                self.nodes_attr[-1]['first_coef'] + self.nodes_attr[-1]['coef_size'])
            attr = dict(inputs=numpy.array(inputs), output=self.size_,
                        coef_size=node.coef.size, first_coef=first_coef)
            self.nodes_attr.append(attr)
        elif len(node.input_weights.shape) == 2:
            if node.input_weights.shape[1] != len(inputs):
                raise RuntimeError(  # pragma: no cover
                    "Dimension mismatch between weights [{}] and inputs [{}].".format(
                        node.input_weights.shape[1], len(inputs)))
            node.nodeid = len(self.nodes)
            self.nodes.append(node)
            first_coef = (
                0 if len(self.nodes_attr) == 0 else
                self.nodes_attr[-1]['first_coef'] + self.nodes_attr[-1]['coef_size'])
            attr = dict(inputs=numpy.array(inputs),
                        output=list(range(self.size_, self.size_ +
                                          node.input_weights.shape[0])),
                        coef_size=node.coef.size, first_coef=first_coef)
            self.nodes_attr.append(attr)
        else:
            raise RuntimeError(  # pragma: no cover
                "Coefficients should have 1 or 2 dimension not {}.".format(node.input_weights.shape))
        self._update_members(node, attr)

    def __getitem__(self, i):
        "Retrieves node and attributes for node i."
        return self.nodes[i], self.nodes_attr[i]

    def __len__(self):
        "Returns the number of nodes"
        return len(self.nodes)

    def _predict_one(self, X):
        res = numpy.zeros((self.size_,), dtype=numpy.float64)
        res[:self.dim] = X
        for node, attr in zip(self.nodes, self.nodes_attr):
            res[attr['output']] = node.predict(res[attr['inputs']])
        return res

    def predict(self, X):
        if len(X.shape) == 2:
            res = numpy.zeros((X.shape[0], self.size_))
            for i, x in enumerate(X):
                res[i, :] = self._predict_one(x)
            return res
        return self._predict_one(X)

    @staticmethod
    def create_from_tree(tree, k=1., arch='one'):
        """
        Creates a @see cl NeuralTreeNet instance from a
        :epkg:`DecisionTreeClassifier`

        @param  tree    :epkg:`DecisionTreeClassifier`
        @param  k       slant of the sigmoïd
        @param  arch    architecture, see below
        @return         @see cl NeuralTreeNet

        The function only works for binary problems.
        Available architecture:
        * `'one'`: the method adds nodes with one output, there
          is no soecific definition of layers,
        * `'compact'`: the adds two nodes, the first computes
          the threshold, the second one computes the leaves
          output, a final node merges all outputs into one

        See notebook :ref:`neuraltreerst` for examples.
        """
        if arch == 'one':
            return NeuralTreeNet._create_from_tree_one(tree, k)
        if arch == 'compact':
            return NeuralTreeNet._create_from_tree_compact(tree, k)
        raise ValueError("Unknown arch value '{}'.".format(arch))

    @staticmethod
    def _create_from_tree_one(tree, k=1.):
        "Implements strategy one. See @see meth create_from_tree."

        if tree.n_classes_ > 2:
            raise RuntimeError(
                "The function only support binary classification problem.")

        n_nodes = tree.tree_.node_count
        children_left = tree.tree_.children_left
        children_right = tree.tree_.children_right
        feature = tree.tree_.feature
        threshold = tree.tree_.threshold
        value = tree.tree_.value.reshape((-1, 2))
        output_class = (value[:, 1] > value[:, 0]).astype(numpy.int64)
        max_features_ = tree.max_features_

        root = NeuralTreeNet(tree.max_features_, empty=True)
        feat_index = numpy.arange(0, max_features_)
        predecessor = {}
        outputs = {i: [] for i in range(0, tree.n_classes_)}
        for i in range(n_nodes):

            if children_left[i] != children_right[i]:
                # node with a threshold
                # right side
                coef = numpy.zeros((max_features_,), dtype=numpy.float64)
                coef[feature[i]] = -k
                node_th = NeuralTreeNode(coef, bias=k * threshold[i],
                                         activation='sigmoid4', tag="N%d-th" % i)
                root.append(node_th, feat_index)

                if i in predecessor:
                    pred = predecessor[i]
                    node1 = pred
                    node2 = node_th
                    attr1 = root[node1.nodeid][1]
                    attr2 = root[node2.nodeid][1]

                    coef = numpy.ones((2,), dtype=numpy.float64) * k
                    node_true = NeuralTreeNode(coef, bias=-k * 1.5,
                                               activation='sigmoid4',
                                               tag="N%d-T" % i)
                    root.append(node_true, [attr1['output'], attr2['output']])

                    coef = numpy.zeros((2,), dtype=numpy.float64)
                    coef[0] = k
                    coef[1] = -k
                    node_false = NeuralTreeNode(coef, bias=-k * 0.25,
                                                activation='sigmoid4',
                                                tag="N%d-F" % i)
                    root.append(node_false, [attr1['output'], attr2['output']])

                    predecessor[children_left[i]] = node_true
                    predecessor[children_right[i]] = node_false
                else:
                    coef = numpy.ones((1,), dtype=numpy.float64) * -1
                    node_false = NeuralTreeNode(
                        coef, bias=1, activation='identity', tag="N%d-F" % i)
                    attr = root[node_th.nodeid][1]
                    root.append(node_false, [attr['output']])

                    predecessor[children_left[i]] = node_th
                    predecessor[children_right[i]] = node_false

            elif i in predecessor:
                # leave
                outputs[output_class[i]].append(predecessor[i])

        # final node
        output = []
        index = [0]
        nb = []
        for i in range(0, tree.n_classes_):
            output.extend(outputs[i])
            nb.append(len(outputs[i]))
            index.append(len(outputs[i]) + index[-1])
        coef = numpy.zeros((len(nb), len(output)), dtype=numpy.float64)
        for i in range(0, tree.n_classes_):
            coef[i, index[i]:index[i + 1]] = k
        feat = [root[n.nodeid][1]['output'] for n in output]
        root.append(
            NeuralTreeNode(coef, bias=(-k / 2) * len(feat),
                           activation='softmax4', tag="Nfinal"),
            feat)

        # final
        return root

    @staticmethod
    def _create_from_tree_compact(tree, k=1.):
        "Implements strategy one. See @see meth create_from_tree."

        if tree.n_classes_ > 2:
            raise RuntimeError(
                "The function only support binary classification problem.")

        n_nodes = tree.tree_.node_count
        children_left = tree.tree_.children_left
        children_right = tree.tree_.children_right
        feature = tree.tree_.feature
        threshold = tree.tree_.threshold
        value = tree.tree_.value.reshape((-1, 2))
        output_class = (value[:, 1] > value[:, 0]).astype(numpy.int64)
        max_features_ = tree.max_features_
        feat_index = numpy.arange(0, max_features_)

        root = NeuralTreeNet(tree.max_features_, empty=True)
        coef1 = []
        bias1 = []
        parents = {}
        rows = {}

        # first pass: threshold

        for i in range(n_nodes):
            if children_left[i] == children_right[i]:
                # leaves
                continue
            rows[i] = len(coef1)
            parents[children_left[i]] = i
            parents[children_right[i]] = i
            coef = numpy.zeros((max_features_,), dtype=numpy.float64)
            coef[feature[i]] = -k
            coef1.append(coef)
            bias1.append(k * threshold[i])

        coef1 = numpy.vstack(coef1)
        if len(bias1) == 1:
            bias1 = bias1[0]
        node1 = NeuralTreeNode(
            coef1 if coef1.shape[0] > 1 else coef1[0], bias=bias1,
            activation='sigmoid4', tag="threshold")
        root.append(node1, feat_index)
        th_index = numpy.arange(max_features_, max_features_ + coef1.shape[0])

        # second pass: decision path
        coef2 = []
        bias2 = []
        output = []

        for i in range(n_nodes):
            if children_left[i] != children_right[i]:
                # not a leave
                continue

            path = []
            last = i
            lr = "class", output_class[i]
            output.append(output_class[i])
            while last is not None:
                path.append((last, lr))
                if last not in parents:
                    break
                par = parents[last]
                if children_right[par] == last:
                    lr = 'right'
                elif children_left[par] == last:
                    lr = 'left'
                else:
                    raise RuntimeError(  # pragma: no cover
                        "Inconsistent tree structure.")
                last = par

            coef = numpy.zeros((coef1.shape[0], ), dtype=numpy.float64)
            bias = 0.
            for ip, lr in path:
                if isinstance(lr, tuple):
                    lr, value = lr
                    if lr != 'class':
                        raise RuntimeError(
                            "algorithm issue")  # pragma: no cover
                else:
                    r = rows[ip]
                    if lr == 'right':
                        coef[r] = k
                        bias -= k / 2
                    else:
                        coef[r] = -k
                        bias += k / 2
            coef2.append(coef)
            bias2.append(bias)

        coef2 = numpy.vstack(coef2)
        if len(bias2) == 1:
            bias2 = bias2[0]
        node2 = NeuralTreeNode(
            coef2 if coef2.shape[0] > 1 else coef2[0], bias=bias2,
            activation='sigmoid4', tag="pathes")
        root.append(node2, th_index)

        # final node
        coef = numpy.zeros(
            (tree.n_classes_, coef2.shape[0]), dtype=numpy.float64)
        bias = [0. for i in range(tree.n_classes_)]
        for i, cls in enumerate(output):
            coef[cls, i] = -k
            coef[1 - cls, i] = k
            bias[cls] += k / 2
            bias[1 - cls] += -k / 2
        findex = numpy.arange(max_features_ + coef1.shape[0],
                              max_features_ + coef1.shape[0] + coef2.shape[0])
        root.append(
            NeuralTreeNode(coef, bias=bias,
                           activation='softmax4', tag="final"),
            findex)

        # end
        return root

    def to_dot(self, X=None):
        """
        Exports the neural network into :epkg:`dot`.

        @param  X   input as an example
        """
        y = None
        if X is not None:
            y = self.predict(X)
        rows = ['digraph Tree {',
                "node [shape=box, fontsize=10];",
                "edge [fontsize=8];"]
        for i in range(self.dim):
            if y is None:
                rows.append('{0} [label="X[{0}]"];'.format(i))
            else:
                rows.append(
                    '{0} [label="X[{0}]=\\n{1:1.2f}"];'.format(i, X[i]))

        labels = {}

        for i in range(0, len(self)):  # pylint: disable=C0200
            o = self[i][1]['output']
            if isinstance(o, int):
                lo = str(o)
                labels[o] = lo
                lof = "%s"
            else:
                lo = "s" + 'a'.join(map(str, o))
                for oo in o:
                    labels[oo] = '{}:f{}'.format(lo, oo)
                los = "|".join("<f{0}> {0}".format(oo) for oo in o)
                lof = "%s&#92;n" + los

            a = "a={}\n".format(self[i][0].activation)
            stag = "" if self[i][0].tag is None else (self[i][0].tag + "\\n")
            bias = str(numpy.array(self[i][0].bias)).replace(" ", "&#92; ")
            if y is None:
                lab = lof % '{}{}id={} b={} s={}'.format(
                    stag, a, i, bias, self[i][0].n_outputs)
            else:
                yo = numpy.array(y[o])
                lab = lof % '{}{}id={} b={} s={}\ny={}'.format(
                    stag, a, i, bias, self[i][0].n_outputs, yo)
            rows.append('{} [label="{}"];'.format(
                lo, lab.replace("\n", "&#92;n")))
            for ii, inp in enumerate(self[i][1]['inputs']):
                if isinstance(o, int):
                    w = self[i][0].input_weights[ii]
                    if w == 0:
                        c = ', color=grey, fontcolor=grey'
                    elif w < 0:
                        c = ', color=red, fontcolor=red'
                    else:
                        c = ', color=blue, fontcolor=blue'
                    rows.append(
                        '{} -> {} [label="{}"{}];'.format(inp, o, w, c))
                    continue

                w = self[i][0].input_weights[:, ii]
                for oi, oo in enumerate(o):
                    if w[oi] == 0:
                        c = ', color=grey, fontcolor=grey'
                    elif w[oi] < 0:
                        c = ', color=red, fontcolor=red'
                    else:
                        c = ', color=blue, fontcolor=blue'
                    rows.append('{} -> {} [label="{}|{}"{}];'.format(
                        labels.get(inp, inp), labels[oo], oi, w[oi], c))

        rows.append('}')
        return '\n'.join(rows)

    @property
    def shape(self):
        "Returns the shape of the coefficients."
        return (sum(n.coef.size for n in self.nodes), )

    @property
    def training_weights(self):
        "Returns the weights."
        sh = self.shape
        res = numpy.empty(sh[0], dtype=numpy.float64)
        pos = 0
        for n in self.nodes:
            s = n.coef.size
            res[pos: pos + s] = (
                n.coef if len(n.coef.shape) == 1 else n.coef.ravel())
            pos += s
        return res

    def update_training_weights(self, X, add=True):  # pylint: disable=W0237
        """
        Updates weights.

        :param grad: vector to add to the weights such as gradient
        :param add: addition or replace
        """
        pos = 0
        if add:
            for n in self.nodes:
                s = n.coef.size
                n.coef += X[pos: pos + s].reshape(n.coef.shape)
                pos += s
        else:
            for n in self.nodes:
                s = n.coef.size
                numpy.copyto(n.coef, X[pos: pos + s].reshape(n.coef.shape))
                pos += s

    def fill_cache(self, X):
        """
        Creates a cache with intermediate results.
        """
        big_cache = {}
        res = numpy.zeros((self.size_,), dtype=numpy.float64)
        res[:self.dim] = X
        for node, attr in zip(self.nodes, self.nodes_attr):
            cache = node.fill_cache(res[attr['inputs']])
            big_cache[node.nodeid] = cache
            res[attr['output']] = cache['aX']
        big_cache[-1] = res
        return big_cache

    def _get_output_node_attr(self, nb_last):
        """
        Retrieves the output nodes.
        *nb_last* is the number of expected outputs.
        """
        neurones = set(self.output_to_node_[i][0].nodeid
                       for i in range(self.size_ - nb_last, self.size_))
        if len(neurones) != 1:
            raise RuntimeError(  # pragma: no cover
                "Only one output node is implemented not {}".format(
                    len(neurones)))
        return self.output_to_node_[self.size_ - 1]

    def _common_loss_dloss(self, X, y, cache=None):
        """
        Common beginning to methods *loss*, *dlossds*,
        *dlossdw*.
        """
        last = 1 if len(y.shape) <= 1 else y.shape[1]
        if cache is not None and -1 in cache:
            res = cache[-1]
        else:
            res = self.predict(X)
        if len(res.shape) == 2:
            pred = res[:, -last:]
        else:
            pred = res[-last:]
        last_node, last_attr = self._get_output_node_attr(last)
        return res, pred, last_node, last_attr

    def loss(self, X, y, cache=None):
        """
        Computes the loss due to prediction error. Returns a float.
        """
        res, _, last_node, last_attr = self._common_loss_dloss(
            X, y, cache=cache)
        if len(res.shape) <= 1:
            return last_node.loss(res[last_attr['inputs']], y)  # pylint: disable=E1120
        return last_node.loss(res[:, last_attr['inputs']], y)  # pylint: disable=E1120

    def dlossds(self, X, y, cache=None):
        """
        Computes the loss derivative against the inputs.
        """
        res, _, last_node, last_attr = self._common_loss_dloss(
            X, y, cache=cache)
        if len(res.shape) <= 1:
            return last_node.dlossds(res[last_attr['inputs']], y)  # pylint: disable=E1120
        return last_node.dlossds(res[:, last_attr['inputs']], y)  # pylint: disable=E1120

    def gradient_backward(self, graddx, X, inputs=False, cache=None):
        """
        Computes the gradient in X.

        :param graddx: existing gradient against the inputs
        :param X: computes the gradient in X
        :param inputs: if False, derivative against the coefficients,
            otherwise against the inputs.
        :param cache: cache intermediate results to avoid more computation
        :return: gradient
        """
        if cache is None:
            cache = self.fill_cache(X)
        shape = self.training_weights.shape
        pred = self.predict(X)

        whole_gradx = numpy.zeros(pred.shape, dtype=numpy.float64)
        whole_gradw = numpy.zeros(shape, dtype=numpy.float64)
        if len(graddx.shape) == 0:
            whole_gradx[-1] = graddx
        else:
            whole_gradx[-graddx.shape[0]:] = graddx

        for node, attr in zip(self.nodes[::-1], self.nodes_attr[::-1]):
            ch = cache[node.nodeid]

            node_graddx = whole_gradx[attr['output']]
            xi = pred[attr['inputs']]

            temp_gradw = node.gradient_backward(
                node_graddx, xi, inputs=False, cache=ch)
            temp_gradx = node.gradient_backward(
                node_graddx, xi, inputs=True, cache=ch)

            whole_gradw[attr['first_coef']:attr['first_coef'] +
                        attr['coef_size']] += temp_gradw.reshape((attr['coef_size'],))
            whole_gradx[attr['inputs']
                        ] += temp_gradx.reshape((len(attr['inputs']),))

        if inputs:
            return whole_gradx
        return whole_gradw



"""
@file
@brief shortcut to image
"""

from .detection_segment import detect_segments, plot_segments, compute_gradient, plot_gradient
from .detection_segment import convert_array2PIL, convert_PIL2array
from .geometrie import Point, Segment
from .queue_binom import tabule_queue_binom
from .random_image import random_noise_image, random_segment_image
